@@ -11,7 +11,7 @@ from typing import Dict
 from typing import List
 from typing import Optional
 
-from falocalrepo_database import FADatabase
+from falocalrepo_database import FADatabase, FADatabaseTable
 from falocalrepo_database import tiered_path
 from flask import Flask
 from flask import abort
@@ -38,17 +38,26 @@ def clean_username(username: str) -> str:
     return str(re_sub(r"[^a-zA-Z0-9\-.~,]", "", username.lower().strip()))
 
 
+def button(href: str, text: str) -> str:
+    return f'<button onclick="window.location = \'{href}\'">{text}</button>'
+
+
 @app.route("/favicon.ico")
 def favicon():
     return redirect("https://www.furaffinity.net/favicon.ico")
 
 
 @app.errorhandler(404)
-def not_found(_err: NotFound):
+def not_found(err: NotFound):
+    return error(err.description, 404)
+
+
+def error(message: str, code: int):
     return render_template(
-        "not_found.html",
-        title=f"{app.name} · Content not Found"
-    ), 404
+        "error.html",
+        title=f"{app.name} · Content not Found",
+        message=f"{code} {message}"
+    ), code
 
 
 @app.route("/")
@@ -87,7 +96,10 @@ def user(username: str):
         user_entry = db.users[username]
 
     if user_entry is None:
-        return abort(404)
+        return error(
+            f"User not found.<br>{button(f'https://www.furaffinity.net/user/{username}', 'Open on Fur Affinity')}",
+            404
+        )
 
     folders: List[int] = list(filter(bool, user_entry["FOLDERS"].split(",")))
     gallery: List[int] = list(map(int, filter(bool, user_entry["GALLERY"].split(","))))
@@ -114,16 +126,6 @@ def search_default():
     return redirect("/search/submissions/")
 
 
-@app.route("/search/submissions/")
-def search_submissions():
-    return search("submissions")
-
-
-@app.route("/search/journals/")
-def search_journals():
-    return search("journals")
-
-
 @app.route("/submissions/<username>/")
 @app.route("/search/submissions/<username>/")
 def search_user_submissions(username: str):
@@ -136,19 +138,49 @@ def search_user_journals(username: str):
     return redirect(f'/search/journals/?author=["{username}"]')
 
 
-def search(table: str):
+@app.route("/search/<string:table>/")
+def search(table: str = "submissions"):
     global last_search
     global db_path
 
-    if request.args:
+    table = table.lower()
+    if table not in ("submissions", "journals", "users"):
+        return error(f"Table {table} not found.", 404)
+
+    with FADatabase(db_path) as db:
+        if not request.args:
+            return render_template(
+                "search.html",
+                title=f"{app.name} · Search {table.title()}",
+                table=table,
+                columns=db[table].columns,
+                params=json_dumps(last_search["params"])
+            )
+
         params: Dict[str, List[str]] = dict(map(
             lambda kv: (kv[0], json_loads(kv[1])),
             request.args.items()
         ))
-        order: List[str] = params.get("order", ["ID DESC"])
+
         limit: int = 50
         offset: int = params.get("offset", 0)
         offset = 0 if offset < 0 else offset
+        columns: List[str]
+        columns_list: List[str]
+        column_id: str
+        order: List[str]
+
+        if table in ("submissions", "journals"):
+            columns = ["ID", "AUTHOR", "TITLE"]
+            columns += ["TAGS"] if table == "submissions" else []
+            columns_list = ["TAGS"] if table == "submissions" else []
+            column_id = "ID"
+            order: List[str] = params.get("order", [f"ID DESC"])
+        elif table == "users":
+            columns = ["USERNAME", "FOLDERS"]
+            columns_list = ["FOLDERS"]
+            column_id = "USERNAME"
+            order: List[str] = params.get("order", [f"USERNAME ASC"])
 
         if "order" in params:
             del params["order"]
@@ -160,40 +192,28 @@ def search(table: str):
         if (last_search["table"], last_search["params"]) != (table, params):
             last_search["table"] = table
             last_search["params"] = deepcopy(params)
+            db_table: FADatabaseTable = db[table]
             if "author" in params:
                 params["replace(author, '_', '')"] = list(map(clean_username, params["author"]))
                 del params["author"]
-            with FADatabase(db_path) as db:
-                if table == "submissions":
-                    if list(params.keys()) == ["order"]:
-                        last_search["results"] = list(db.submissions)
-                    else:
-                        last_search["results"] = list(db.submissions.cursor_to_dict(
-                            db.submissions.select(params, like=True, order=order)))
-                elif table == "journals":
-                    if list(params.keys()) == ["order"]:
-                        last_search["results"] = list(db.journals)
-                    else:
-                        last_search["results"] = list(db.journals.cursor_to_dict(
-                            db.journals.select(params, like=True, order=order)))
+            if list(params.keys()) == ["order"]:
+                last_search["results"] = list(db_table)
+            else:
+                last_search["results"] = list(db_table.cursor_to_dict(db_table.select(params, like=True, order=order)))
 
         return render_template(
             "search_results.html",
             title=f"{app.name} · {table.title()} Search Results",
             table=table,
             params=last_search["params"],
+            columns=columns,
+            columns_list=columns_list,
+            column_id=column_id,
             limit=limit,
             offset=offset,
             results=last_search["results"][offset:offset + limit],
             results_total=len(last_search["results"])
         )
-    else:
-        return render_template(
-            "search.html",
-            title=f"{app.name} · Search {table.title()}",
-            table=table,
-            params=json_dumps(last_search["params"])
-        )
 
 
 @app.route("/journal/<int:id_>/")
@@ -205,7 +225,10 @@ def journal(id_: int):
         jrnl = db.journals[id_]
 
     if jrnl is None:
-        return abort(404)
+        return error(
+            f"Journal not found.<br>{button(f'https://www.furaffinity.net/journal/{id_}', 'Open on Fur Affinity')}",
+            404
+        )
 
     return render_template(
         "journal.html",
@@ -228,7 +251,10 @@ def submission(id_: int):
         sub: Optional[dict] = db.submissions[id_]
 
     if sub is None:
-        return abort(404)
+        return error(
+            f"Submission not found.<br>{button(f'https://www.furaffinity.net/view/{id_}', 'Open on Fur Affinity')}",
+            404
+        )
 
     file_type: Optional[str] = ""
     if (ext := sub["FILEEXT"]) in ("jpg", "jpeg", "png", "gif"):
